"""Labware definition for the custom Copan 24x rack.
This file can be
 - imported: e.g. `from copan_24 import Copan24Specs`.
    Labware definition can be accessed with the method Copan24Specs.labware_definition
 - executed with opentrons_simulate or opetrons_execute: e.g. `opentrons_simulate covmatic_stations/a/copan_24.py`
    This file acts as a test protocol for the custom labware.
 - executed with python:  e.g. `python -m covmatic_stations.a.copan_24`.
    This script generates the json file for the custom labware"""
import json
from collections import OrderedDict
from itertools import product, chain
from typing import List, Tuple
from opentrons_shared_data.labware.dev_types import LabwareDefinition
from opentrons.protocol_api import ProtocolContext


class JsonProperty(property):
    _count = 0
    
    def __init__(self, fget=None, fset=None, fdel=None, doc=None, ord=0):
        super(JsonProperty, self).__init__(fget=fget, fset=fset, fdel=fdel, doc=doc)
        self._idx = type(self)._count
        type(self)._count += 1
    
    def __int__(self) -> int:
        return self._idx


json_property = JsonProperty


class Copan24Specs:
    def __init__(self,
                 nrows: int = 4,
                 ncols: int = 6,
                 brand: str = "COPAN",
                 brandId: List[str] = ["330C"],
                 diameter: float = 16.8,
                 distance_vert: float = 17.75,
                 distance_horz: float = 18.75,
                 tube_diameter: float = 16,
                 tube_volume: float = 14000,
                 tube_depth: float = 93,
                 tube_height: float = 10,
                 global_dimensions: Tuple[float, float, float] = (121.35, 79.1, 108),
                 a1_offset: Tuple[float, float] = (11.63, 14),
                 ):
        self.nrows = nrows
        self.ncols = ncols
        self._brand = brand
        self._brandId = brandId
        self._dims = global_dimensions
        self._a1_off = a1_offset
        self._d = diameter
        self._dv = distance_vert
        self._dh = distance_horz
        self._tw = tube_diameter
        self._td = tube_depth
        self._tv = tube_volume
        self._th = tube_height
    
    @property
    def n(self) -> int:
        return self.ncols * self.nrows

    @json_property
    def ordering(self) -> List[List[str]]:
        return [[chr(r + ord("A")) + str(c + 1) for r in range(self.nrows)] for c in range(self.ncols)]

    @json_property
    def brand(self) -> dict:
        return {"brand": self._brand, "brandId": self._brandId}

    @json_property
    def metadata(self) -> dict:
        return {
            "displayName": "COPAN {} Tube Rack 14000 µL".format(self.n),
            "displayCategory": "tubeRack",
            "displayVolumeUnits": "µL",
            "tags": []
        }

    @json_property
    def dimensions(self) -> dict:
        return {
            "xDimension": self._dims[0],
            "yDimension": self._dims[1],
            "zDimension": self._dims[2]
        }
    
    def well(self, r: int, c: int) -> Tuple[str, dict]:
        return chr(ord("A") + r) + str(1 + c), {
            "depth": self._td,
            "totalLiquidVolume": self._tv,
            "shape": "circular",
            "diameter": self._tw,
            "x": self._a1_off[0] + c * self._dh,
            "y": self._dims[1] - self._a1_off[1] - r * self._dv,
            "z": self._th
        }
        
    @json_property
    def wells(self) -> dict:
        return dict(self.well(r, c) for c, r in product(range(self.ncols), range(self.nrows)))
    
    @json_property
    def groups(self) -> List[dict]:
        d = {k: self.metadata[k] for k in ("displayName", "displayCategory")}
        d["wellBottomShape"] = "v"
        return [{
            "metadata": d,
            "brand": self.brand,
            "wells": list(chain.from_iterable(self.ordering)),
        }]

    @json_property
    def parameters(self) -> dict:
        return {
            "format": "irregular",
            "quirks": [],
            "isTiprack": False,
            "isMagneticModuleCompatible": False,
            "loadName": "copan_{}_tuberack_14000ul".format(self.n)
        }

    @json_property
    def namespace(self) -> str:
        return "custom_beta"

    @json_property
    def version(self) -> int:
        return 1

    @json_property
    def schemaVersion(self) -> int:
        return 2

    @json_property
    def cornerOffsetFromSlot(self) -> dict:
        return {
            "x": 0,
            "y": 0,
            "z": 0
        }
    
    def toJSON(self) -> dict:
        return OrderedDict((k, getattr(self, k)) for k in sorted(
            filter(lambda a: isinstance(getattr(type(self), a, None), json_property), dir(self)),
            key=lambda a: int(getattr(type(self), a, 0))
        ))
    
    def labware_definition(self) -> LabwareDefinition:
        return LabwareDefinition(self.toJSON())
    
    def __str__(self) -> str:
        return json.dumps(self.toJSON(), indent=4).replace(r"\u00b5", "\u00b5")
    
    def run_test(self, ctx: ProtocolContext):
        """Test protocol"""
        ctx.comment("Test the custom '{}' rack".format(self.metadata["displayName"]))
        
        rack = ctx.load_labware_from_definition(self.labware_definition(), '2', 'custom tuberack')
        tipracks1000 = [ctx.load_labware('opentrons_96_filtertiprack_1000ul', '1', '1000µl filter tiprack')]
        p1000 = ctx.load_instrument('p1000_single_gen2', 'right', tip_racks=tipracks1000)
        
        p1000.pick_up_tip()
        for w in rack.wells():
            ctx.pause("moving to top of {}".format(w))
            p1000.move_to(w.top())
            ctx.pause("moving to bottom of {} (1 mm high)".format(w))
            p1000.move_to(w.bottom(1))
            p1000.aspirate(5)
            p1000.dispense(5)
        p1000.drop_tip()
        

def run(ctx: ProtocolContext):
    Copan24Specs().run_test(ctx)


metadata = {"apiLevel": "2.3"}


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('file', metavar='F', type=str, help='The file path where to save the custom labware JSON')
    args = parser.parse_args()
    with open(args.file, "w") as f:
        f.write(str(Copan24Specs()))


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from ..station import Station, labware_loader, instrument_loader
from ..utils import mix_bottom_top, uniform_divide, mix_walk
from . import magnets
from opentrons.types import Point
from typing import Optional, Tuple
import logging
import math


class StationB(Station):
    _protocol_description = "station B protocol"
    
    def __init__(
        self,
        bind_air_gap: float = 20,
        bind_aspiration_rate: float = 50,
        bind_blowout_rate: float = 300,
        bind_dispense_rate: float = 150,
        bind_max_transfer_vol: float = 180,
        bind_mix_loc_bottom: float = 1,
        bind_mix_loc_top: float = 5,
        bind_mix_times: int = 8,
        bind_mix_vol: float = 180,
        bind_sample_mix_times: int = 5,
        bind_sample_mix_vol: float = 20,
        bind_vol: float = 210,
        bottom_headroom_height: float = 0.5,
        default_aspiration_rate: float = 150,
        drop_height: float = 0,
        drop_loc_l: float = -10,
        drop_loc_r: float = 30,
        drop_loc_y: float = 0,
        drop_threshold: int = 296,
        elute_air_gap: float = 20,
        elute_aspiration_rate: float = 50,
        elute_incubate: bool = True,
        elute_mix_times: int = 10,
        elute_mix_vol: float = 30,
        elution_height: float = 5,
        elution_vol: float = 40,
        jupyter: bool = True,
        logger: Optional[logging.getLoggerClass()] = None,
        h_trans: float = 10,
        h_bottom: float = 0.5,
        n_bottom: int = 3,
        vol_last_trans: float = 100, 
        magdeck_slot: str = '1',
        magheight: float = 6.65,
        magheight_load: bool = True,
        magplate_model: str = 'nest_96_wellplate_2ml_deep',
        metadata: Optional[dict] = None,
        num_samples: int = 96,
        samples_per_col: int = 8,
        skip_delay: bool = False,
        supernatant_removal_air_gap: float = 5,
        supernatant_removal_aspiration_rate: float = 25,
        supernatant_removal_height: float = 0.5,
        supernatant_removal_last_transfer_max_vol: float = 190,
        starting_vol: float = 380,
        tempdeck_slot: str = '1',
        tempdeck_temp: float = 4,
        tipracks_slots: Tuple[str, ...] = ('3', '6', '7', '8', '9', '10'),
        touch_tip_height: float = -5,
        wait_time_bind_off: float = 5,
        wait_time_bind_on: float = 5,
        wait_time_dry: float = 12,
        wait_time_elute_off: float = 5,
        wait_time_elute_on: float = 3,
        wait_time_wash_on: float = 5,
        wash_air_gap: float = 20,
        wash_etoh_times: int = 4,
        wash_etoh_vol: float = 800,
        wash_headroom: float = 1.1,
        wash_max_transfer_vol: float = 200,
        wash_mix_speed: float = 20,
        wash_mix_vol: float = 150,
        wash_mix_walk: bool = False,
        wash_1_times: int = 20,
        wash_1_vol: float = 500,
        wash_2_times: int = 20,
        wash_2_vol: float = 500,
        wash_2_mix_aspiration_rate: float = 150,
        wash_2_mix_dispense_rate: float = 160,
        wash_1_mix_aspiration_rate: float = 120,
        wash_1_mix_dispense_rate: float = 120,
        **kwargs
    ):
        """ Build a :py:class:`.StationB`.
        :param bind_air_gap: Air gap for bind beads in uL
        :param bind_aspiration_rate: Aspiration flow rate when aspirating bind beads in uL/s
        :param bind_blowout_rate: Blowout flow rate when aspirating bind beads in uL/s
        :param bind_dispense_rate: Dispensation flow rate when aspirating bind beads in uL/s
        :param bind_max_transfer_vol: Maximum volume transferred of bind beads
        :param bind_mix_loc_bottom: Mixing location for bind beads at the bottom in mm
        :param bind_mix_loc_top: Mixing location for bind beads at the top in mm
        :param bind_mix_times: Mixing repetitions for bind beads
        :param bind_mix_vol: Mixing volume for bind beads
        :param bind_sample_mix_times: Mixing repetitions for bind beads and samples
        :param bind_sample_mix_vol: Mixing repetitions for bind beads in uL
        :param bind_vol: Total volume transferred of bind beads
        :param bottom_headroom_height: Height to keep from the bottom
        :param default_aspiration_rate: Default aspiration flow rate in uL/s
        :param drop_loc_l: offset for dropping to the left side (should be positive) in mm
        :param drop_loc_r: offset for dropping to the right side (should be negative) in mm
        :param drop_loc_y: offset in the fron/back direction for dropping in mm
        :param drop_threshold: the amount of dropped tips after which the run is paused for emptying the trash
        :param elute_air_gap: Air gap when aspirating elution buffer in uL
        :param elute_aspiration_rate: Aspiration flow rate when aspirating elution buffer in uL/s
        :param elute_incubate: Whether or not to wait for incubation between elution mix and transfer (if True, magdeck will also activate)
        :param elute_mix_times: Mix times for elution
        :param elute_mix_vol: Mix volume for elution
        :param elution_height: Height at which to sample after elution in mm
        :param elution_vol: The volume of elution buffer to aspirate in uL
        :param logger: logger object. If not specified, the default logger is used that logs through the ProtocolContext comment method
        :param h_trans: total liquid height expected starting on B
        :param h_bottom: height offset from the bottom 
        :param n_bottom: number of aspirating phases done in the last phase 
        :param vol_last_trans: volume left for the last section phase
        :param magheight: Height of the magnet, in mm
        :param magheight_load: Load magheight from JSON, by serial (if no serial number is found, fall back onto magheight parameter)
        :param magplate_model: Magnetic plate model
        :param metadata: protocol metadata
        :param num_samples: The number of samples that will be loaded on the station B
        :param samples_per_col: The number of samples in a column of the destination plate
        :param skip_delay: If True, pause instead of delay.
        :param supernatant_removal_air_gap: Air gap when removing the supernatant in uL
        :param supernatant_removal_aspiration_rate: Aspiration flow rate when removing the supernatant in uL/s
        :param supernatant_removal_height: Height from the bottom when removing the supernatant in mm
        :param supernatant_removal_last_transfer_max_vol: maximum volume to load into tip during supernatant last transfer
        :param starting_vol: Sample volume at start (volume coming from Station A)
        :param tempdeck_slot: Slot where the tempdeck is positioned 
        :param tempdeck_temp: tempdeck temperature in Celsius degrees 
        :param tipracks_slots: Slots where the tipracks are positioned
        :param touch_tip_height: Touch-tip height in mm (should be negative)
        :param wait_time_bind_off: Wait time for bind beads phase off magnet in minutes
        :param wait_time_bind_on: Wait time for bind beads phase on magnet in minutes
        :param wait_time_dry: Wait time for airdrying phase in minutes
        :param wait_time_elute_off: Wait time for elution phase off magnet in minutes
        :param wait_time_elute_on: Wait time for elution phase on magnet in minutes
        :param wait_time_wash_on: Wait time for wash phase on magnet in minutes
        :param wash_air_gap: Air gap for wash in uL
        :param wash_etoh_times: Mix times for ethanol
        :param wash_etoh_vol: Volume of ethanol in uL
        :param wash_headroom: Headroom for wash buffers (as a multiplier)
        :param wash_max_transfer_vol: Maximum volume transferred of wash in uL
        :param wash_mix_aspiration_rate: Aspiration flow rate when mixing wash buffer in uL/s
        :param wash_mix_dispense_rate: Dispensation flow rate when mixing wash buffer in uL/s
        :param wash_mix_speed: Movement speed of the pipette while mixing in mm/s
        :param wash_mix_vol: Mix volume for wash
        :param wash_mix_walk: Whether to move or not when mixing the wash buffer
        :param wash_1_times: Mix times for wash 1
        :param wash_1_vol: Volume of wash 1 buffer in uL
        :param wash_2_times: Mix times for wash 2
        :param wash_2_vol: Volume of wash 2 buffer in uL
        """
        super(StationB, self).__init__(
            drop_loc_l=drop_loc_l,
            drop_loc_r=drop_loc_r,
            drop_loc_y=drop_loc_y,
            jupyter=jupyter,
            logger=logger,
            metadata=metadata,
            num_samples=num_samples,
            samples_per_col=samples_per_col,
            skip_delay=skip_delay,
            magdeck_slot=magdeck_slot,
            wash_2_mix_aspiration_rate=wash_2_mix_aspiration_rate,
            wash_2_mix_dispense_rate=wash_2_mix_dispense_rate,
            wash_1_mix_aspiration_rate=wash_1_mix_aspiration_rate,
            wash_1_mix_dispense_rate=wash_1_mix_dispense_rate,
            **kwargs
        )
        self._bind_air_gap = bind_air_gap
        self._bind_aspiration_rate = bind_aspiration_rate
        self._bind_blowout_rate = bind_blowout_rate
        self._bind_dispense_rate = bind_dispense_rate
        self._bind_max_transfer_vol = bind_max_transfer_vol
        self._bind_mix_loc_bottom = bind_mix_loc_bottom
        self._bind_mix_loc_top = bind_mix_loc_top
        self._bind_mix_times = bind_mix_times
        self._bind_mix_vol = bind_mix_vol
        self._bind_sample_mix_times = bind_sample_mix_times
        self._bind_sample_mix_vol = bind_sample_mix_vol
        self._bind_vol = bind_vol
        self._bottom_headroom_height = bottom_headroom_height
        self._default_aspiration_rate = default_aspiration_rate
        self._drop_height = drop_height
        self._drop_threshold = drop_threshold
        self._elute_air_gap = elute_air_gap
        self._elute_aspiration_rate = elute_aspiration_rate
        self._elute_incubate = elute_incubate
        self._elute_mix_times = elute_mix_times
        self._elute_mix_vol = elute_mix_vol
        self._elution_height = elution_height
        self._elution_vol = elution_vol
        self._h_trans = h_trans
        self._h_bottom = h_bottom
        self._n_bottom = n_bottom
        self._vol_last_trans = vol_last_trans
        self._magdeck_slot = magdeck_slot
        self._magheight = magheight
        self._magheight_load = magheight_load
        self._magplate_model = magplate_model
        self._supernatant_removal_air_gap = supernatant_removal_air_gap
        self._supernatant_removal_aspiration_rate = supernatant_removal_aspiration_rate
        self._supernatant_removal_height = supernatant_removal_height
        self._supernatant_removal_last_transfer_max_vol = supernatant_removal_last_transfer_max_vol
        self._starting_vol = starting_vol
        self._tempdeck_slot = tempdeck_slot
        self._tempdeck_temp = tempdeck_temp
        self._tipracks_slots = tipracks_slots
        self._touch_tip_height = touch_tip_height
        self._wait_time_bind_off = wait_time_bind_off
        self._wait_time_bind_on = wait_time_bind_on
        self._wait_time_dry = wait_time_dry
        self._wait_time_elute_off = wait_time_elute_off
        self._wait_time_elute_on = wait_time_elute_on
        self._wait_time_wash_on = wait_time_wash_on
        self._wash_air_gap = wash_air_gap
        self._wash_etoh_times = wash_etoh_times
        self._wash_etoh_vol = wash_etoh_vol
        self._wash_headroom = wash_headroom
        self._wash_max_transfer_vol = wash_max_transfer_vol
        self._wash_2_mix_aspiration_rate = wash_2_mix_aspiration_rate
        self._wash_2_mix_dispense_rate = wash_2_mix_dispense_rate
        self._wash_1_mix_aspiration_rate = wash_1_mix_aspiration_rate
        self._wash_1_mix_dispense_rate = wash_1_mix_dispense_rate
        self._wash_mix_speed = wash_mix_speed
        self._wash_mix_vol = wash_mix_vol
        self._wash_mix_walk = wash_mix_walk
        self._wash_1_times = wash_1_times
        self._wash_1_vol = wash_1_vol
        self._wash_2_times = wash_2_times
        self._wash_2_vol = wash_2_vol
    
    @labware_loader(0, "_tips300")
    def load_tips300(self):
        self._tips300 = [
            self._ctx.load_labware('opentrons_96_tiprack_300ul', slot, '200µl filtertiprack')
            for slot in self._tipracks_slots
        ]
    
    @labware_loader(2, "_magdeck")
    def load_magdeck(self):
        self._magdeck = self._ctx.load_module('Magnetic Module Gen2', self._magdeck_slot)
        self._magdeck.disengage()
        if (self._magheight_load):
            mag_deck_serial = self._magdeck._module._driver.get_device_info()['serial']
            self._magheight = magnets.height.by_serial.get(mag_deck_serial, self._magheight)
            self.logger.info("Found MagDeck with serial {}. Engage height is {}mm".format(mag_deck_serial, self._magheight))
    
    @labware_loader(3, "_magplate")
    def load_magplate(self):
        self._magplate = self._magdeck.load_labware(self._magplate_model)
        self.logger.debug("using '{}' magnetic plate".format(self._magplate_model))
    
    @property
    def mag_samples_m(self):
        return self._magplate.rows()[0][:self.num_cols]
    
    @labware_loader(4, "_tempdeck")
    def load_tempdeck(self):
        self._tempdeck = self._ctx.load_module('Temperature Module Gen2', self._tempdeck_slot)
        if self._tempdeck_temp is not None:
            self.msg = "set temperature"
            self._tempdeck.set_temperature(self._tempdeck_temp)
    
    @labware_loader(5, "_flatplate")
    def load_flatplate(self):
        self._flatplate = self._tempdeck.load_labware('opentrons_96_aluminumblock_nest_wellplate_100ul')
    
    @property
    def elution_samples_m(self):
        return self._flatplate.rows()[0][:self.num_cols]
    
    @labware_loader(6, "_waste")
    def load_waste(self):
        self._waste = self._ctx.load_labware('nest_1_reservoir_195ml', '11', 'Liquid Waste').wells()[0].top()
    
    @labware_loader(7, "_etoh")
    def load_etoh(self):
        self._etoh = self._ctx.load_labware('nest_1_reservoir_195ml', '2', 'Trough with Ethanol').wells()[:1]
    
    @labware_loader(8, "_res12")
    def load_res12(self):
        self._res12 = self._ctx.load_labware('nest_12_reservoir_15ml', '5', 'Trough with WashReagents')
    
    @property
    def binding_buffer(self):
        return self._res12.wells()[:2]
    
    @property
    def wash1(self):
        return self._res12.wells()[3:7]
    
    @property
    def wash2(self):
        return self._res12.wells()[7:11]
    
    @property
    def water(self):
        """Elution reagent"""
        return self._res12.wells()[11]
    
    @instrument_loader(0, "_m300")
    def load_m300(self):
        self._m300 = self._ctx.load_instrument('p300_multi_gen2', 'left', tip_racks=self._tips300)
        if self._bind_aspiration_rate:
            self._m300.flow_rate.aspirate = self._bind_aspiration_rate
        if self._bind_dispense_rate:
            self._m300.flow_rate.dispense = self._bind_dispense_rate
        if self._bind_blowout_rate:
            self._m300.flow_rate.blow_out = self._bind_blowout_rate
    
    def _tipracks(self) -> dict:
        return {"_tips300": "_m300",}

    def check(self):
        self._ctx.comment("Check module status")
        for j in range(3):
            if self._magdeck.status == 'disengaged':
                self._magdeck.engage(height=self._magheight)
                if self._magdeck.status == 'engaged':
                    self._ctx.comment("Engaged module: {} attempt".format(j+1))
        if self._magdeck.status == 'disengaged':
            self.dual_pause("check module")

    def remove_supernatant(self, vol: float, stage: str = "remove supernatant"):
        self._m300.flow_rate.aspirate = self._supernatant_removal_aspiration_rate
        
        num_trans = math.ceil((vol - self._vol_last_trans) / self._bind_max_transfer_vol)
        vol_per_trans = (vol - self._vol_last_trans) / num_trans
        # h_num_trans = self._h_trans/num_trans

        for i, m in enumerate(self.mag_samples_m):
            if self.run_stage("{} {}/{}".format(stage, i + 1, len(self.mag_samples_m))):
                self.pick_up(self._m300)
                loc = m.bottom(self._supernatant_removal_height)
                self._ctx.comment("Supernatant removal: {} transfer with {}ul each.".format(num_trans, vol_per_trans))
                for _ in range(num_trans):
                    # num_trans = num_trans - 1
                    # aspirate_height = self.h_bottom + (num_trans * h_num_trans) # expecting aspirated height
                    # ctx.comment('Aspirating at {}'.format(aspirate_height))
                    if self._m300.current_volume > 0:
                        self._m300.dispense(self._m300.current_volume, m.top())                   
                    self._m300.move_to(m.center())
                    self._m300.transfer(vol_per_trans, loc, self._waste, new_tip='never', air_gap=self._supernatant_removal_air_gap)
                    self._m300.air_gap(self._supernatant_removal_air_gap)

                self._ctx.comment("Supernatant removal: last transfer in {} step".format(self._n_bottom))
                # dispensing the air gap present in the tip
                if self._m300.current_volume > 0:
                    self._m300.dispense(self._m300.current_volume, m.top())
                #self._m300.move_to(m.center())         ## Dovrebbe essere già fatto da m.bottom() seguente

                for j in range(self._n_bottom):
                    aspirate_height = self._h_bottom - (j+1)*(self._h_bottom/self._n_bottom) # expecting aspirated height
                    loc = m.bottom(aspirate_height)
                    self._m300.aspirate(self._supernatant_removal_last_transfer_max_vol/self._n_bottom, loc)
                    self._ctx.comment("Aspirating at {}".format(aspirate_height))

                self._m300.air_gap(self._supernatant_removal_air_gap)
                self._m300.dispense(self._m300.current_volume, self._waste)
                self._m300.air_gap(self._supernatant_removal_air_gap)
                self.drop(self._m300)
        self._m300.flow_rate.aspirate = self._default_aspiration_rate
        
    def bind(self):
        """Add bead binding buffer and mix samples"""
        self._m300.flow_rate.aspirate = self._bind_aspiration_rate
        
        for i, well in enumerate(self.mag_samples_m):
            if self.run_stage("transfer binding {}/{}".format(i + 1, len(self.mag_samples_m))):
                source = self.binding_buffer[i // ((len(self.mag_samples_m) // len(self.binding_buffer)) or 1)]
                self.pick_up(self._m300)
                mix_bottom_top(
                    self._m300,
                    self._bind_mix_times,
                    self._bind_mix_vol,
                    source.bottom,
                    self._bind_mix_loc_bottom,
                    self._bind_mix_loc_top
                )
                
                num_trans, vol_per_trans = uniform_divide(self._bind_vol, self._bind_max_transfer_vol)
                
                for t in range(num_trans):
                    if self._m300.current_volume > 0:
                        self._m300.dispense(self._m300.current_volume, source.top())  # void air gap if necessary
                    self._m300.transfer(vol_per_trans, source, well.top(), air_gap=self._bind_air_gap, new_tip='never')
                    if t == 0:
                        self._m300.air_gap(self._bind_air_gap)
                self._m300.mix(self._bind_sample_mix_times, self._bind_sample_mix_vol, well)
                
                self._m300.touch_tip(v_offset=self._touch_tip_height)
                self._m300.air_gap(self._bind_air_gap)
                self.drop(self._m300)
        
        if self.run_stage("bind wait"):
            # Time Issue in Station B After the waiting time of 5 min the magnetic module should run for 6 min.
            self.delay(self._wait_time_bind_off, 'magnet wait')
        self._magdeck.engage(height=self._magheight)
        self.check()
        
        if self.run_stage("bind incubate"):
            # Time Issue in Station B After the waiting time of 5 min the magnetic module should run for 6 min.
            self.delay(self._wait_time_bind_on, self.get_msg_format("incubate on magdeck", self.get_msg("on")))

        # Remove initial supernatant
        self.remove_supernatant(self._bind_vol + self._starting_vol, "remove binding")
    
    @staticmethod
    def wash_getcol(sample_col_idx: int, wash_cols: int, source):
        return source[sample_col_idx // ((wash_cols // len(source)) or 1)]
    
    def wash(self, vol: float, source, mix_reps: int, wash_name: str = "wash"):
        self.logger.info(self.msg_format("wash info", vol, wash_name, mix_reps))
        if wash_name == "wash 1":
            self._default_aspiration_rate = self._wash_1_mix_aspiration_rate
            dispense_rate = self._wash_1_mix_dispense_rate
            self._m300.flow_rate.dispense = self._wash_1_mix_dispense_rate
            self._m300.flow_rate.aspirate = self._wash_1_mix_aspiration_rate
        else:
            self._default_aspiration_rate = self._wash_2_mix_aspiration_rate
            dispense_rate = self._wash_2_mix_dispense_rate
            self._m300.flow_rate.dispense = self._wash_2_mix_dispense_rate
            self._m300.flow_rate.aspirate = self._wash_2_mix_aspiration_rate
        self._magdeck.disengage()
        num_trans, vol_per_trans = uniform_divide(vol, self._wash_max_transfer_vol)
        
        for i, m in enumerate(self.mag_samples_m):
            if self.run_stage("{} {}/{}".format(wash_name, i + 1, len(self.mag_samples_m))):
                self.pick_up(self._m300)
                src = self.wash_getcol(i, len(self.mag_samples_m), source)
                
                for n in range(num_trans):
                    if self._m300.current_volume > 0:
                        self._m300.dispense(self._m300.current_volume, src.top())
                    self._m300.transfer(vol_per_trans, src, m.top(), air_gap=20, new_tip='never')
                    if n < num_trans - 1:  # only air_gap if going back to source
                        self._m300.air_gap(self._wash_air_gap)
                
                # Mix
                if self._wash_mix_walk:
                    a_locs = [m.bottom(self._bottom_headroom_height).move(Point(x=2*(-1 if i % 2 else +1), y=2*(2*j/(mix_reps - 1) - 1))) for j in range(mix_reps)]
                    mix_walk(self._m300, mix_reps, self._wash_mix_vol, a_locs, speed=self._wash_mix_speed, logger=self.logger)
                else:
                    loc = m.bottom(self._bottom_headroom_height).move(Point(x=2*(-1 if i % 2 else +1)))
                    self._m300.mix(mix_reps, self._wash_mix_vol, loc)
                self._m300.flow_rate.aspirate = self._default_aspiration_rate
                self._m300.flow_rate.dispense = dispense_rate
                
                self._m300.air_gap(self._wash_air_gap)
                self.drop(self._m300)

        self._magdeck.engage(height=self._magheight)
        self.check()
        if self.run_stage("{} incubate".format(wash_name)):
            self.delay(self._wait_time_wash_on, self.get_msg_format("incubate on magdeck", self.get_msg("on")))
        self.remove_supernatant(vol, stage="remove {}".format(wash_name))
    
    def elute(self, positions=None, transfer: bool = True, stage: str = "elute"):
        """Resuspend beads in elution"""
        if positions is None:
            positions = self.mag_samples_m
        self._m300.flow_rate.aspirate = self._elute_aspiration_rate
        for i, m in enumerate(positions):
            if self.run_stage("{} {}/{}".format(stage, i + 1, len(positions))):
                self.pick_up(self._m300)
                side = 1 if i % 2 == 0 else -1
                loc = m.bottom(self._bottom_headroom_height).move(Point(x=side*2))
                self._m300.aspirate(self._elution_vol, self.water)
                self._m300.air_gap(self._elute_air_gap)
                self._m300.dispense(self._elute_air_gap, m.top())
                self._m300.dispense(self._elution_vol, loc)
                self._m300.mix(self._elute_mix_times, self._elute_mix_vol, loc)
                self._m300.touch_tip(v_offset=self._touch_tip_height)
                self._m300.air_gap(self._elute_air_gap)
                self.drop(self._m300)
        
        if self._elute_incubate and self.run_stage("{} incubate off".format(stage)):
            self.delay(self._wait_time_elute_off, self.get_msg_format("incubate on magdeck", self.get_msg("off")))
        self._magdeck.engage(height=self._magheight)
        self.check()
        if self._elute_incubate and self.run_stage("{} incubate on".format(stage)):
            self.delay(self._wait_time_elute_on, self.get_msg_format("incubate on magdeck", self.get_msg("on")))
        
        if transfer:
            for i, (m, e) in enumerate(zip(
                positions,
                self.elution_samples_m
            )):
                if self.run_stage("{} transfer {}/{}".format(stage, i + 1, len(positions))):
                    self.pick_up(self._m300)
                    side = -1 if i % 2 == 0 else 1
                    loc = m.bottom(self._bottom_headroom_height).move(Point(x=side*2))
                    self._m300.transfer(self._elution_vol, loc, e.bottom(self._elution_height), air_gap=self._elute_air_gap, new_tip='never')
                    # m300.blow_out(e.top(-2))
                    self._m300.air_gap(self._elute_air_gap)
                    self.drop(self._m300)
    
    def body(self):
        self.bind()
        self.wash(self._wash_1_vol, self.wash1, self._wash_1_times, "wash 1")
        self.wash(self._wash_2_vol, self.wash2, self._wash_2_times, "wash 2")
        self.wash(self._wash_etoh_vol, self._etoh, self._wash_etoh_times, "ethanol")
        self._magdeck.disengage()
        if self.run_stage("airdry beads"):
            self.delay(self._wait_time_dry, 'airdry')
        self.elute()
        self._magdeck.disengage()


if __name__ == "__main__":
    StationB(metadata={'apiLevel': '2.3'}).simulate()


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from opentrons.protocol_api import ProtocolContext
from threading import Thread
from functools import wraps
import requests
import time


class Dummyable(type):
    class Dummy: pass
    
    def __new__(mcs, classname: str, supers: tuple, classdict: dict):
        cls = super(Dummyable, mcs).__new__(mcs, classname, supers, classdict)
        if mcs.Dummy not in cls.__mro__:
            def emptyfun(*args, **kwargs): pass
            dummydict = {k: v if k[:2] == "__" else (wraps(v)(emptyfun) if callable(v) else None) for k, v in map(lambda k: (k, getattr(cls, k, None)), dir(cls))}
            cls.dummy = type("Dummy{}".format(classname), (mcs.Dummy, cls), dummydict)
        return cls


class BlinkingLight(Thread, metaclass=Dummyable):
    def __init__(self, ctx: ProtocolContext, t: float = 1):
        super(BlinkingLight, self).__init__()
        self._on = False
        self._state = True
        self._state_initial = None
        self._ctx = ctx
        self._t = t
    
    def stop(self):
        self._on = False
        self.join()
    
    def initial_state(self) -> bool:
        return self._ctx._hw_manager.hardware.get_lights()
    
    def set_light(self, s: bool):
        self._ctx._hw_manager.hardware.set_lights(rails=s)
    
    def run(self):
        self._on = True
        self._state_initial = self.initial_state()
        while self._on:
            self._state = not self._state
            self.set_light(self._state)
            time.sleep(self._t)
        self.set_light(self._state_initial)


class BlinkingLightHTTP(BlinkingLight):
    _URL = "http://127.0.0.1:31950/robot/lights"
    _OT_request_headers = {"Opentrons-Version":  "2"}

    def initial_state(self) -> bool:
        return requests.get(self._URL, headers=self._OT_request_headers).json().get('on', False)
    
    def set_light(self, s: bool):
        requests.post(self._URL, headers=self._OT_request_headers, json={'on': s})


class Button(metaclass=Dummyable):
    _base_cols = ['red', 'green', 'blue']
    _all_cols = ['black', 'blue', 'green', 'cyan', 'red', 'magenta', 'yellow', 'white']
    
    def __init__(self, ctx: ProtocolContext, color: str = 'blue'):
        self._ctx = ctx
        self._default_color = color
        self.color = color

    @classmethod
    def decode(cls, state: dict) -> str:
        return cls._all_cols[sum(1 << i for i, c in enumerate(reversed(cls._base_cols)) if state[c])]

    @classmethod
    def encode(cls, color: str) -> dict:
        idx = cls._all_cols.index(color) if color in cls._all_cols else 0
        return dict(zip(cls._base_cols, map(bool, map(int, bin(idx)[2:5].zfill(3)))))
    
    @property
    def color(self) -> str:
        return self.decode(self._state)

    @color.setter
    def color(self, color: str):
        self._state = self.encode(color)
        self._ctx._hw_manager.hardware._backend.gpio_chardev.set_button_light(**self._state)
    
    def __del__(self):
        self.color = self._default_color


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



"""Labware definition for the custom Copan 48x rack.
This file can be
 - imported: e.g. `from copan_48 import StaggeredCopan48Specs`.
    Labware definition can be accessed with the method StaggeredCopan48Specs.labware_definition
 - executed with opentrons_simulate or opetrons_execute: e.g. `opentrons_simulate covmatic_stations/a/copan_48.py`
    This file acts as a test protocol for the custom labware.
 - executed with python:  e.g. `python -m covmatic_stations.a.copan_48`.
    This script generates the json file for the custom labware"""
from covmatic_stations.a.copan_24 import Copan24Specs, json_property
from opentrons.protocol_api import ProtocolContext
from typing import Tuple
import inspect
import copy
import os
import json


_a1_offset = (27.5, 12)
_global_dimensions = (260, 177, 118)
_nrows = 8
_ncols = 6


class StaggeredCopan48Specs(Copan24Specs):
    def __init__(
        self,
        stagger: float = 0.33,
        nrows: int = _nrows,
        ncols: int = _ncols,
        global_dimensions=_global_dimensions,
        distance_horz: float = (_global_dimensions[0] - 2 * _a1_offset[0])/(_ncols - 1),
        distance_vert: float = (_global_dimensions[1] - 2 * _a1_offset[1])/(_nrows - 1),
        a1_offset=_a1_offset,
        **kwargs
    ):
        super(StaggeredCopan48Specs, self).__init__(
            nrows=nrows,
            ncols=ncols,
            global_dimensions=global_dimensions,
            distance_horz=distance_horz,
            distance_vert=distance_vert,
            a1_offset=a1_offset,
            **kwargs
        )
        self._stagger = stagger
    
    @json_property
    def metadata(self) -> dict:
        d = super(StaggeredCopan48Specs, self).metadata
        d["displayName"] = "COPAN {} Staggered Tube Rack 14000 µL".format(self.n)
        return d
        
    def well(self, r: int, c: int) -> Tuple[str, dict]:
        w = super(StaggeredCopan48Specs, self).well(r, c)
        w[1]["x"] += (+1 if r % 2 else -1) * self._tw * self._stagger
        return w


class StaggeredCopan48SpecsCorrected(StaggeredCopan48Specs):
    def __init__(self, **kwargs):
        remaining_kwargs = copy.deepcopy(kwargs)
        corrected_args = {}
        for k, v in inspect.signature(super(StaggeredCopan48SpecsCorrected, self).__init__).parameters.items():
            if k in remaining_kwargs:
                remaining_kwargs.pop(k)
                if hasattr(kwargs[k], "__iter__"):
                    corrected_args[k] = tuple(a * d for a, d in zip(kwargs[k], v.default))
                else:
                    corrected_args[k] = kwargs[k] * v.default
        super(StaggeredCopan48SpecsCorrected, self).__init__(**corrected_args, **remaining_kwargs)


copan_48_correction_env_key = "OT_COPAN_48_CORRECT"
copan_48_correction_file = os.path.join(os.path.dirname(__file__), "copan_48_correction.json")
copan_48_correction_file = os.environ.get(copan_48_correction_env_key, copan_48_correction_file)
with open(copan_48_correction_file, "r") as f:
    copan_48_correction = json.load(f)


copan_48_corrected_specs = StaggeredCopan48SpecsCorrected(**copan_48_correction)


def run(ctx: ProtocolContext):
    StaggeredCopan48Specs().run_test(ctx)


metadata = {"apiLevel": "2.3"}


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('file', metavar='F', type=str, help='The file path where to save the custom labware JSON')
    args = parser.parse_args()
    with open(args.file, "w") as f:
        f.write(str(StaggeredCopan48Specs()))


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from ..station import instrument_loader, labware_loader
from .c import StationC
import math
import copy
from itertools import chain, repeat
from typing import Tuple


_MM_MIX = {
    "a": 6,
    "b": 8,
    "c": 6,
}


class StationCTechnogenetics(StationC):
    _protocol_description = "station C protocol for Technogenetics kit"
    
    def __init__(
        self,
        mm_mix: dict = _MM_MIX,
        mm_strip_capacity: float = 180,
        mm_tube_capacity: float = 1800,
        pause_on_mastermix_msg: bool = True,
        source_plate_name: str = 'chilled elution plate on block for Station B',
        tempdeck_bool: bool = False,
        tiprack_slots: Tuple[str, ...] = ('2', '3', '5', '6', '8', '9', '11'),
        transfer_samples: bool = False,
        tube_block_model: str = "opentrons_24_aluminumblock_nest_2ml_screwcap",
        **kwargs
    ):
        """ Build a :py:class:`.StationCTechnogenetics`.
        :param mm_mix: Mastermix reagent quantities per sample in uL
        :param mm_strip_capacity: Capacity of one cell of the strip for mastermix in uL
        :param mm_tube_capacity: Capacity of one tube for mastermix in uL
        :param pause_on_mastermix_msg: Pause when diplaying message with mastermix composition
        :param kwargs: other keyword arguments. See: StationC, Station
        """
        super(StationCTechnogenetics, self).__init__(
            source_plate_name=source_plate_name,
            tipracks_slots=tiprack_slots,
            transfer_samples=transfer_samples,
            tube_block_model=tube_block_model,
            **kwargs
        )
        self._mm_mix = copy.deepcopy(mm_mix)
        self._mm_strip_capacity = mm_strip_capacity
        self._mm_tube_capacity = mm_tube_capacity
        self._pause_on_mastermix_msg = pause_on_mastermix_msg
        self._tempdeck_bool = tempdeck_bool
        self._transfer_samples = transfer_samples
    
    def _tipracks(self) -> dict:
        return {
            "_tips300": "_p300",
            "_tips20": "_m20",
        }
    
    def load_tips20_no_a(self): pass
    
    @property
    def mm_per_sample(self) -> float:
        return sum(self._mm_mix.values())
    
    @property
    def _mastermix_vol(self) -> float:
        return self.mm_per_sample
    
    @_mastermix_vol.setter
    def _mastermix_vol(self, ignored):
        pass
    
    @property
    def mm_capacity(self) -> float:
        return min(self._mm_tube_capacity, 8 * self._mm_strip_capacity)
    
    @property
    def num_mm_tubes(self) -> int:
        return int(math.ceil(self.mm_per_sample * self._samples_this_cycle * self._mastermix_vol_headroom / self.mm_capacity))
    
    @property
    def samples_per_mm_tube(self) -> Tuple[int, ...]:
        samples_per_mm_tube = []
        for i in range(self.num_mm_tubes):
            remaining_samples = self._samples_this_cycle - sum(samples_per_mm_tube)
            samples_per_mm_tube.append(min(8 * int(math.ceil(remaining_samples / (8 * (self.num_mm_tubes - i)))), remaining_samples))
        return tuple(samples_per_mm_tube)
    
    @property
    def mm_per_tube(self) -> Tuple[float, ...]:
        return tuple(self.mm_per_sample * self._mastermix_vol_headroom * ns for ns in self.samples_per_mm_tube)
    
    @property
    def mm_tubes(self):
        return self._tube_block.wells()[:self.num_mm_tubes]
    
    @property
    def mm_strips(self):
        return self._mm_strips.columns()[:self.num_mm_tubes]
    
    @property
    def mm_indices(self):
        return list(chain.from_iterable(repeat(i, ns) for i, ns in enumerate(self.samples_per_mm_tube)))
    
    def log_mm_mix_info(self) -> str:
        ndigs = math.ceil(math.log10(math.floor(max(self.mm_per_tube) + 1)))
        fmt = lambda n: ("{:>" + str(ndigs + 3) + "}").format("{:.2f}".format(n))
        msg = ""
        for i, (mt, mm, ns) in enumerate(zip(self.mm_tubes, self.mm_per_tube, self.samples_per_mm_tube)):
            msg += (
                "\n  {} --> {} uL".format(str(mt).split(" ")[0], fmt(mm)) +
                "".join("\n    {} -> {} uL".format(k, fmt(ns * v * self._mastermix_vol_headroom)) for k, v in self._mm_mix.items())
            )
        return self.get_msg_format("load tubes", self.num_mm_tubes, msg)
    
    def cycle_begin(self):
        super(StationCTechnogenetics, self).cycle_begin()
        self.logger.debug("samples this cycle {}".format(self._samples_this_cycle))
        self.logger.debug("num mm tubes {}".format(self.num_mm_tubes))
        msg = self.log_mm_mix_info()
        if self._pause_on_mastermix_msg and self.run_stage("mastermix info{}{}".format(" " if self.num_cycles > 1 else "", self._cycle)):
            self.dual_pause(msg, home=(False, False))
        else:
            for r in msg.split("\n"):
                while "  " in r:
                    r = r.replace("  ", "\u2007 ")
                self.logger.info(r)


class StationCTechnogeneticsM300(StationCTechnogenetics):
    # variable names are kept as before for easy inheritance
    # although pipette is now a m300 
    @labware_loader(1, "_tips20")
    def load_tips20(self):
        self._tips20 = [
            self._ctx.load_labware('opentrons_96_filtertiprack_200ul', slot)
            for slot in self._tipracks_slots
        ]
    
    @instrument_loader(0, "_m20")
    def load_m20(self):
        self._m20 = self._ctx.load_instrument('p300_multi_gen2', 'right', tip_racks=self._tips20)
        self._m20.flow_rate.aspirate = self._aspirate_rate
        self._m20.flow_rate.dispense = self._dispense_rate



from ..station import Station, labware_loader, instrument_loader
from ..geometry import LysisTube
from ..utils import mix_bottom_top
from itertools import chain, islice, repeat
import math
import logging
from typing import Optional, Tuple


class StationA(Station):    
    def __init__(
        self,
        air_gap_dest_multi: float = 5,
        air_gap_sample: float = 20,
        air_gap_lys: float = 10,
        dest_headroom_height: float = 2,
        dest_top_height: float = 5,
        dest_multi_headroom_height: float = 2,
        drop_loc_l: float = -10,
        drop_loc_r: float = 30,
        drop_threshold: int = 296,
        hover_height: float = -2,
        ic_capacity: float = 180,
        ic_lys_headroom: float = 1.1,
        ic_headroom_bottom: float = 2,
        ic_mix_repeats: int = 5,
        ic_mix_volume: float = 20,
        iec_volume: float = 19,
        internal_control_idx_th: int = 48,
        jupyter: bool = True,
        logger: Optional[logging.getLoggerClass()] = None,
        lysis_cone_height: float = 16,
        lysis_first: bool = False,
        lysis_headroom_height: float = 4,
        lysis_rate_aspirate: float = 100,
        lysis_rate_dispense: float = 100,
        lysis_volume: float = 160,
        lys_mix_repeats: int = 10,
        lys_mix_volume: float = 100,
        main_pipette: str = 'p300_single_gen2',
        main_tiprack: str = 'opentrons_96_tiprack_300ul',
        main_tiprack_label: str = '200ul filter tiprack',
        max_speeds_a: float = 20,
        metadata: dict = None,
        mix_repeats: int = 3,
        mix_volume: float = 150,
        num_samples: int = 96,
        positive_control_well: str = 'A10',
        sample_aspirate: float = 30,
        sample_blow_out: float = 300,
        sample_dispense: float = 30,
        samples_per_col: int = 8,
        sample_volume: float = 200,
        source_headroom_height: float = 8,
        source_position_top: float = -10,
        source_top_height: float = 10,
        source_racks: str = 'opentrons_24_tuberack_nest_1.5ml_screwcap',
        source_racks_definition_filepath: str = "",
        source_racks_slots: Tuple[str, ...] = ('2', '3', '5', '6'),
        tempdeck_temp: float = 4,
        tempdeck_bool: bool = True,
        tipracks_slots: Tuple[str, ...] = ('8', '9', '11'),
        tipracks_slots_20: Tuple[str, ...] = ('7',),
        **kwargs
    ):
        """ Build a :py:class:`.StationA`.

        :param air_gap_dest_multi: air gap for destination tube (multi) in uL
        :param air_gap_sample: air gap for sample transfer in uL
        :param dest_headroom_height: headroom always to keep from the bottom of the destination tube in mm
        :param dest_top_height: top height from the bottom of the destination tube in mm when mixing
        :param dest_multi_headroom_height: headroom always to keep from the bottom of the destination tube (multi) in mm
        :param drop_loc_l: offset for dropping to the left side (should be positive) in mm
        :param drop_loc_r: offset for dropping to the right side (should be negative) in mm
        :param drop_threshold: the amount of dropped tips after which the run is paused for emptying the trash
        :param hover_height: height from the top at which to hover (should be negative) in mm
        :param ic_capacity: capacity of the internal control tube
        :param ic_lys_headroom: headroom for the internal control and lysis tube (multiplier)
        :param ic_headroom_bottom: headroom always to keep from the bottom of the internal control tube in mm
        :param ic_mix_repeats: number of repetitions during mixing the internal control
        :param ic_mix_volume: volume aspirated for mixing the internal control in uL
        :param iec_volume: The volume of lysis buffer to use per sample in uL
        :param internal_control_idx_th: internal control index threshold for choosing the strip 
        :param logger: logger object. If not specified, the default logger is used that logs through the ProtocolContext comment method
        :param lysis_cone_height: height of he conic bottom of the lysis buffer tube in mm
        :param lysis_first: whether to transfer the lysis buffer first or else the sample first
        :param lysis_headroom_height: headroom always to keep from the bottom of the lysis buffer tube in mm
        :param lysis_rate_aspirate: P300 aspiration flow rate when aspirating lysis buffer in uL/s
        :param lysis_rate_dispense: P300 dispensation flow rate when dispensing lysis buffer in uL/s
        :param lysis_volume: The volume of lysis buffer to use per sample in uL
        :param lys_mix_repeats: number of repetitions during mixing the lysis buffer
        :param lys_mix_volume: volume aspirated for mixing the lysis buffer in uL
        :param main_pipette: type of the main pipette
        :param main_tiprack: type of the main tiprack
        :param main_tiprack_label: label of the main tiprack
        :param mix_repeats: number of repetitions during mixing
        :param mix_volume: volume aspirated for mixing in uL
        :param max_speeds_a: max speed for sample transfer
        :param metadata: protocol metadata
        :param num_samples: The number of samples that will be loaded on the station A
        :param positive_control_well: Position of the positive control well
        :param sample_aspirate: P300 samples aspiration flow rate in uL/s
        :param sample_blow_out: P300 samples blow out flow rate in uL/s
        :param sample_dispense: P300 samples dispensation flow rate in uL/s
        :param samples_per_col: The number of samples in a column of the destination plate
        :param sample_volume: The volume of a sample in uL
        :param source_headroom_height: headroom always to keep from the bottom of the source tube in mm
        :param source_position_top: height from the top of the source tube in mm (should be negative)
        :param source_top_height: top height from the bottom of the source tube in mm when mixing
        :param source_racks: source racks to load
        :param source_racks_definition_filepath: filepath for source racks definitions
        :param source_racks_slots: slots where source racks are installed
        :param tempdeck_temp: tempdeck temperature in Celsius degrees
        :param tipracks_slots: Slots where the tipracks are positioned
        :param tipracks_slots_20: Slots where the tipracks (20 uL) are positioned
        :param jupyter: Specify whether the protocol is run on Jupyter (or Python) instead of the robot
        """
        super(StationA, self).__init__(
            drop_loc_l=drop_loc_l,
            drop_loc_r=drop_loc_r,
            drop_threshold=drop_threshold,
            jupyter=jupyter,
            logger=logger,
            metadata=metadata,
            num_samples=num_samples,
            samples_per_col=samples_per_col,
            **kwargs
        )
        self._air_gap_dest_multi = air_gap_dest_multi
        self._air_gap_sample = air_gap_sample
        self._air_gap_lys = air_gap_lys
        self._dest_headroom_height = dest_headroom_height
        self._dest_multi_headroom_height = dest_multi_headroom_height
        self._dest_top_height = dest_top_height
        self._hover_height = hover_height
        self._ic_capacity = ic_capacity
        self._ic_headroom_bottom = ic_headroom_bottom
        self._ic_lys_headroom = ic_lys_headroom
        self._ic_mix_repeats = ic_mix_repeats
        self._ic_mix_volume = ic_mix_volume
        self._iec_volume = iec_volume
        self._internal_control_idx_th = internal_control_idx_th
        self._lysis_cone_height = lysis_cone_height
        self._lysis_first = lysis_first
        self._lysis_headroom_height = lysis_headroom_height
        self._lysis_rate_aspirate = lysis_rate_aspirate
        self._lysis_rate_dispense = lysis_rate_dispense
        self._lysis_volume = lysis_volume
        self._lys_mix_repeats = lys_mix_repeats
        self._lys_mix_volume = lys_mix_volume
        self._main_pipette = main_pipette
        self._main_tiprack = main_tiprack
        self._main_tiprack_label = main_tiprack_label
        self._max_speeds_a = max_speeds_a
        self._mix_repeats = mix_repeats
        self._mix_volume = mix_volume
        self._positive_control_well = positive_control_well
        self._sample_aspirate = sample_aspirate
        self._sample_blow_out = sample_blow_out
        self._sample_dispense = sample_dispense
        self._sample_volume = sample_volume
        self._source_headroom_height = source_headroom_height
        self._source_position_top = source_position_top
        self._source_racks = source_racks
        self._source_racks_definition_filepath = source_racks_definition_filepath
        self._source_racks_slots = source_racks_slots
        self._source_top_height = source_top_height
        self._tempdeck_temp = tempdeck_temp
        self._tempdeck_bool = tempdeck_bool
        self._tipracks_slots = tipracks_slots
        self._tipracks_slots_20 = tipracks_slots_20
    
    @labware_loader(0, "_tempdeck")
    def load_tempdeck(self):
        if self._tempdeck_bool:
            self._tempdeck = self._ctx.load_module('Temperature Module Gen2', '4')
            if self._tempdeck_temp is not None:
                self._tempdeck.set_temperature(self._tempdeck_temp)
        else:
            pass
    
    @property
    def chilled_tubeblock_content(self) -> str:
        return self.get_msg_format("chilled tubeblock content", self.num_ic_strips, "{}")
    
    @labware_loader(0.1, "_strips_block")
    def load_strips_block(self):
        if self._tempdeck_bool:
            self._strips_block = self._tempdeck.load_labware(
                'opentrons_96_aluminumblock_generic_pcr_strip_200ul',
                "chilled tubeblock for {}".format(self.chilled_tubeblock_content.format(""))
            )
        else:
            self._strips_block = self._ctx.load_labware('opentrons_96_aluminumblock_generic_pcr_strip_200ul', '4',
                "chilled tubeblock for {}".format(self.chilled_tubeblock_content.format("")))

        self.logger.info("{} {}".format(
            type(self).get_message("using", self._language),
            self.chilled_tubeblock_content.format(self.get_msg_format("chilled tubeblock content with", self.cols_per_strip * self._iec_volume * self._ic_lys_headroom))
        ))
    
    def _load_source_racks(self):
        self._source_racks = [
            self._ctx.load_labware(
                self._source_racks, slot,
                'source tuberack ' + str(i + 1)
            ) for i, slot in enumerate(self._source_racks_slots)
        ]
    
    @labware_loader(1, "_source_racks")
    def load_source_racks(self):
        self.logger.debug("using source racks '{}'".format(self._source_racks))
        self._load_source_racks()
    
    @labware_loader(2, "_dest_plate")
    def load_dest_plate(self):
        self._dest_plate = self._ctx.load_labware(
            'nest_96_wellplate_2ml_deep', '1', '96-deepwell sample plate'
        )
    
    _lys_buf_name: str = '50ml tuberack for lysis buffer'
    
    @labware_loader(3, "_lys_buff")
    def load_lys_buf(self):
        self._lys_buff = self._ctx.load_labware(
            'opentrons_6_tuberack_falcon_50ml_conical', '7',
            self._lys_buf_name,
        ).wells()[0]
    
    @labware_loader(4, "_tipracks_main")
    def load_tipracks_main(self):
        self._tipracks_main = [
            self._ctx.load_labware(self._main_tiprack, slot, self._main_tiprack_label)
            for slot in self._tipracks_slots
        ]
        self.logger.debug("main tipracks: {}".format(", ".join(map(str, self._tipracks_main))))
    
    @labware_loader(5, "_tipracks20")
    def load_tipracks20(self):
        self._tipracks20 = [
            self._ctx.load_labware('opentrons_96_filtertiprack_20ul', slot, '20ul filter tiprack')
            for slot in self._tipracks_slots_20
        ]
    
    @instrument_loader(0, "_m20")
    def load_m20(self):
        self._m20 = self._ctx.load_instrument('p20_multi_gen2', 'left', tip_racks=self._tipracks20)
    
    @instrument_loader(1, "_p_main")
    def load_p_main(self):
        self._p_main = self._ctx.load_instrument(self._main_pipette, 'right', tip_racks=self._tipracks_main)
        self._p_main.flow_rate.blow_out = self._sample_blow_out
        self.logger.debug("main pipette: {}".format(self._p_main))
    
    @property
    def num_ic_strips(self) -> int:
        return math.ceil(self._iec_volume * self.num_cols * self._ic_lys_headroom / self._ic_capacity)
    
    @property
    def cols_per_strip(self) -> int:
        return math.ceil(self.num_cols / self.num_ic_strips)
    
    @property
    def initlial_volume_lys(self) -> float:
        return len(list(self.non_control_positions(repeat(None)))) * self._lysis_volume * self._ic_lys_headroom

    def setup_samples(self):
        self._sources = list(islice(chain.from_iterable(rack.wells() for rack in self._source_racks), self._num_samples))
        self._dests_single = self._dest_plate.wells()[:self._num_samples]
        self._dests_multi = self._dest_plate.rows()[0][:self.num_cols]
        self.logger.debug("positive control in {} of destination rack".format(self._positive_control_well))
    
    def setup_lys_tube(self):
        self._lysis_tube = LysisTube(self._lys_buff.diameter / 2, self._lysis_cone_height)
        self.logger.debug("%s", self._lysis_tube)
        self._lysis_tube.height = self._lysis_headroom_height
        self._lysis_tube.fill(self.initlial_volume_lys)
        self._lysis_tube.height += self._ic_lys_headroom
        self.logger.info(self.msg_format("lysis geometry", math.ceil(self._lysis_tube.volume), self._lysis_tube.height))
        self._lysis_tube.height -= self._ic_lys_headroom

    def transfer_sample(self, source, dest):
        self.logger.debug("transferring from {} to {}".format(source, dest))
        self.pick_up(self._p_main)
        
        self._p_main.move_to(source.top(self._source_position_top))
        self._ctx.max_speeds['A'] = self._max_speeds_a
        
        # Mix by aspirating and dispensing at different heights
        mix_bottom_top(
            self._p_main, self._mix_repeats, self._mix_volume,
            source.bottom, self._source_headroom_height, self._source_top_height
        )
        self._p_main.aspirate(self._sample_volume, source.bottom(self._source_headroom_height))        
        
        # Wait to be sure the sample is aspirated
        self._ctx.delay(1)
        self._p_main.move_to(source.top(self._source_position_top))
        self._ctx.max_speeds['A'] = None
        self._p_main.air_gap(self._air_gap_sample)
        
        self._p_main.dispense(self._air_gap_sample, dest.top(self._hover_height))
        self._p_main.dispense(self._sample_volume, dest.bottom(self._dest_top_height))
        
        if self._lysis_first:
            self._p_main.flow_rate.aspirate = self._lysis_rate_aspirate
            self._p_main.flow_rate.dispense = self._lysis_rate_dispense
            
            # Mix with lysis buffer
            mix_bottom_top(
                self._p_main, self._lys_mix_repeats, self._lys_mix_volume,
                dest.bottom, self._dest_headroom_height, self._dest_top_height
            )
            
            self._p_main.flow_rate.aspirate = self._sample_aspirate
            self._p_main.flow_rate.dispense = self._sample_dispense
            
        self._p_main.air_gap(self._air_gap_sample)
        self.drop(self._p_main)
    
    def is_positive_control_well(self, dest) -> bool:
        return dest == self._dest_plate[self._positive_control_well]
    
    def non_control_positions(self, sources=None, dests=None):
        """Returns the iterator for the source/dest couples,
        excluding the couple where the destination is meant for the positive control.
        Sources and dests default to self._sources and self._dests"""
        sources = sources or self._sources
        dests = dests or self._dests_single
        return filter(lambda t: not self.is_positive_control_well(t[1]), zip(sources, dests))

    def transfer_samples(self):
        self._p_main.flow_rate.aspirate = self._sample_aspirate
        self._p_main.flow_rate.dispense = self._sample_dispense

        n = len(list(self.non_control_positions()))
        for i, (s, d) in enumerate(self.non_control_positions()):
            if self.run_stage("transfer sample {}/{}".format(i + 1, n)):
                self.transfer_sample(s, d)

    def transfer_lys(self):
        self._p_main.flow_rate.aspirate = self._lysis_rate_aspirate
        self._p_main.flow_rate.dispense = self._lysis_rate_dispense

        if self._lysis_first:
            self.pick_up(self._p_main)
        mix = {} if self._lysis_first else {'mix_after': (self._lys_mix_repeats, self._lys_mix_volume)}
        pos = list(self.non_control_positions(repeat(None)))
        n = len(pos)
        for i, (_, dest) in enumerate(pos):
            h = max(
                self._lysis_tube.extract(self._lysis_volume * self._ic_lys_headroom),
                self._lysis_headroom_height
            )
            if self.run_stage("transfer lysis {}/{}".format(i + 1, n)):
                if not self._lysis_first:
                    self.pick_up(self._p_main)
                self.logger.debug("transferring lysis to {}".format(dest))
                self.logger.debug("going {} mm deep".format(h))
                self._p_main.transfer(
                    self._lysis_volume,
                    self._lys_buff.bottom(h),
                    dest.bottom(self._lysis_headroom_height),
                    air_gap=self._air_gap_lys,
                    new_tip='never',
                    **mix
                )
                self._p_main.air_gap(self._air_gap_sample)
                if self._lysis_first:
                    self._p_main.dispense(self._air_gap_sample, self._lys_buff.top())
                else:
                    self.drop(self._p_main)
        if self._lysis_first:
            self.drop(self._p_main)
    
    def transfer_internal_control(self, idx: int, dest):
        self._p_main.flow_rate.aspirate = self._lysis_rate_aspirate
        self._p_main.flow_rate.dispense = self._lysis_rate_dispense
        
        strip_ind = idx // self.cols_per_strip
        self.logger.debug("transferring internal control strip {}/{} to {}".format(strip_ind + 1, self.num_ic_strips, dest))
        internal_control = self._strips_block.rows()[0][strip_ind]
        self.pick_up(self._m20)
        # no air gap to use 1 transfer only avoiding drop during multiple transfers
        self._m20.transfer(self._iec_volume, internal_control, dest.bottom(self._ic_headroom_bottom), new_tip='never')
        self._m20.mix(self._ic_mix_repeats, self._ic_mix_volume, dest.bottom(self._dest_multi_headroom_height))
        self._m20.air_gap(self._air_gap_dest_multi)
        self.drop(self._m20)
    
    def transfer_internal_controls(self):
        n = len(self._dests_multi)
        for i, d in enumerate(self._dests_multi):
            if self.run_stage("transfer internal control {}/{}".format(i + 1, n)):
                self.transfer_internal_control(i, d)
    
    def _tipracks(self) -> dict:
        return {
            "_tipracks_main": "_p_main",
            "_tipracks20": "_m20",
        }
    
    def body(self):
        self.setup_samples()
        self.setup_lys_tube()
        self.msg = ""
        
        T = (self.transfer_lys, self.transfer_samples)
        for t in (T if self._lysis_first else reversed(T)):
            t()
        
        self.dual_pause("incubate", between=self.set_external)
        self.set_internal()
        self.transfer_internal_controls()
        self.logger.info(self.msg_format("move to B"))


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from opentrons.protocol_api import ProtocolContext
from opentrons.types import Location
import logging
import math
import requests
from itertools import tee, cycle, islice, chain, repeat
from typing import Tuple, Union, Iterable, Callable, Optional, Dict, Any


class ProtocolContextLoggingHandler(logging.Handler):
    """Logging Handler that emits logs through the ProtocolContext comment method"""
    def __init__(self, ctx: ProtocolContext, *args, **kwargs):
        super(ProtocolContextLoggingHandler, self).__init__(*args, **kwargs)
        self._ctx = ctx
    
    def emit(self, record):
        try:
            self._ctx.comment(self.format(record))
        except Exception:
            self.handleError(record)


class LocalWebServerLogger:
    def __init__(self, ip: Optional[str] = None, endpoint: str = ":5002/log", timeout: float = 2, *args, **kwargs):
        super(LocalWebServerLogger, self).__init__(*args, **kwargs)
        self.ip = ip
        self.endpoint = endpoint
        self.timeout = timeout
        self.level = 0
        self.last_dollar = None
    
    @property
    def url(self) -> str:
        return "http://{}{}".format(self.ip, self.endpoint)
    
    def format(self, record):
        if self.last_dollar == record['$']:
            if record['$'] == 'before':
                self.level += 1
            else:
                self.level -= 1
        self.last_dollar = record['$']
        if record['$'] == 'before':
            return ' '.join(['\t' * self.level, record['payload'].get('text', '').format(**record['payload'])])
    
    def __call__(self, record: Dict[str, Any]):
        s = self.format(record)
        url = self.url
        if url and s:
            try:
                requests.post(
                    url,
                    s.encode('utf-8'),
                    headers={'Content-type': 'text/plain; charset=utf-8'},
                    timeout=self.timeout,
                )
            except Exception:
                pass


def mix_bottom_top(pip, reps: int, vol: float, pos: Callable[[float], Location], bottom: float, top: float):
    """Custom mixing procedure aspirating at the bottom and dispensing at the top
    :param pip: The pipette
    :param reps: Number of repeats
    :param vol: Volume to mix
    :param pos: Method for getting the position
    :param bottom: Offset for the bottom position
    :param top: Offset for the top position"""
    for _ in range(reps):
        pip.aspirate(vol, pos(bottom))
        pip.dispense(vol, pos(top))


def mix_walk(
    pip,
    reps: int,
    vol: float,
    aspirate_locs: Union[Iterable, Location],
    dispense_locs: Optional[Union[Iterable, Location]] = None,
    speed: Optional[float] = None,
    logger: Optional[logging.getLoggerClass()] = None
):
    """Custom mixing procedure aspirating and dispensing at custom positions
    :param pip: The pipette
    :param reps: Number of repeats
    :param vol: Volume to mix
    :param aspirate_locs: Position(s) at which to aspirate. If less than reps, they are cycled over.
    :param dispense_locs: Position(s) at which to dispense (optional). If less than reps, they are cycled over. If not specified, dispense in place
    :param speed: Speed for moving the pipette around the mixing position (optional). At the end, the previous speed is restored
    :param logger: Logger for debugging information (optional)"""
    if isinstance(aspirate_locs, Location):
        aspirate_locs = [aspirate_locs]
    if dispense_locs is None:
        aspirate_locs, dispense_locs = tee(aspirate_locs, 2)
    elif isinstance(dispense_locs, Location):
        dispense_locs = [dispense_locs]
    
    old_speed = pip.default_speed
    
    for b, a, d in islice(zip(chain([True], repeat(False)), cycle(aspirate_locs), cycle(dispense_locs)), reps):
        if b and speed is not None:
            pip.move_to(a)
            pip.default_speed = speed
            if logger is not None:
                logger.debug('set speed to {}'.format(speed))
        if logger is not None:
            logger.debug('mixing at {} and {}'.format(a, d))
        pip.aspirate(vol, a)
        pip.dispense(vol, d)
    
    if logger is not None and speed is not None:
        logger.debug('set speed to {}'.format(old_speed))
    pip.default_speed = old_speed


def uniform_divide(total: float, mpp: float) -> Tuple[int, float]:
    """Return the minimum number of partitions and the quantity per partition that uniformly divide a given quantity
    :param total: The total quantity to divide
    :param mpp: Maximum quantity per partition
    :returns: The minimum number of partitions and the quantity in each partition"""
    n = int(math.ceil(total / mpp))
    p = total / n
    return n, p


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from ..station import Station, labware_loader, instrument_loader
from itertools import chain
import math
import logging
from itertools import repeat
from typing import Optional, Tuple


class StationC(Station):
    _protocol_description = "station C protocol"
    
    def __init__(
        self,
        aspirate_rate: float = 30,
        bottom_headroom_height: float = 0.5,
        dispense_rate: float = 30,
        drop_loc_l: float = 0,
        drop_loc_r: float = 0,
        drop_threshold: int = 296,
        jupyter: bool = True,
        logger: Optional[logging.getLoggerClass()] = None,
        mastermix_vol: float = 12,
        mastermix_vol_headroom: float = 1.2,
        mastermix_vol_headroom_aspirate: float = 20/18,
        metadata: Optional[dict] = None,
        num_samples: int = 96,
        positive_control_well: str = 'A10',
        sample_blow_height: float = -2,
        sample_bottom_height: float = 2,
        sample_mix_vol: float = 10,
        sample_mix_reps: int = 1,
        sample_vol: float = 8,
        samples_per_col: int = 8,
        samples_per_cycle: int = 96,
        skip_delay: bool = False,
        source_plate_name: str = 'chilled elution plate on block from Station B',
        suck_height: float = 2,
        suck_vol: float = 5,
        tempdeck_bool: bool = True,
        tipracks_slots: Tuple[str, ...] = ('2', '3', '6', '7', '9'),
        transfer_samples: bool = True,
        tube_block_model: str = "opentrons_24_aluminumblock_nest_1.5ml_snapcap",
        **kwargs
    ):
        """ Build a :py:class:`.StationC`.
        :param aspirate_rate: Aspiration rate in uL/s
        :param bottom_headroom_height: Height to keep from the bottom
        :param dispense_rate: Dispensation rate in uL/s
        :param drop_loc_l: offset for dropping to the left side (should be positive) in mm
        :param drop_loc_r: offset for dropping to the right side (should be negative) in mm
        :param drop_threshold: the amount of dropped tips after which the run is paused for emptying the trash
        :param logger: logger object. If not specified, the default logger is used that logs through the ProtocolContext comment method
        :param mastermix_vol: Mastermix volume per sample in uL
        :param mastermix_vol_headroom: Headroom for mastermix preparation volume as a multiplier
        :param mastermix_vol_headroom_aspirate: Headroom for mastermix aspiration volume as a divisor
        :param metadata: protocol metadata
        :param num_samples: The number of samples that will be loaded on the station B
        :param positive_control_well: Position of the positive control well
        :param sample_blow_height: Height from the top when blowing out in mm (should be negative)
        :param sample_bottom_height: Height to keep from the bottom in mm when dealing with samples
        :param sample_mix_vol: Samples mixing volume in uL
        :param sample_mix_reps: Samples mixing repetitions 
        :param sample_vol: Sample volume
        :param samples_per_col: The number of samples in a column of the destination plate
        :param samples_per_cycle: The number of samples processable in one cycle
        :param source_plate_name: Name for the source plate
        :param skip_delay: If True, pause instead of delay.
        :param suck_height: Height from the top when sucking in any remaining droplets on way to trash in mm
        :param suck_vol: Volume for sucking in any remaining droplets on way to trash in uL
        :param transfer_samples: Whether to transfer samples or not
        :param tube_block_model: Tube block model name
        """
        super(StationC, self).__init__(
            drop_loc_l=drop_loc_l,
            drop_loc_r=drop_loc_r,
            drop_threshold=drop_threshold,
            jupyter=jupyter,
            logger=logger,
            metadata=metadata,
            num_samples=num_samples,
            samples_per_col=samples_per_col,
            skip_delay=skip_delay,
            **kwargs
        )
        self._aspirate_rate = aspirate_rate
        self._bottom_headroom_height = bottom_headroom_height
        self._dispense_rate = dispense_rate
        self._mastermix_vol = mastermix_vol
        self._mastermix_vol_headroom = mastermix_vol_headroom
        self._mastermix_vol_headroom_aspirate = mastermix_vol_headroom_aspirate
        self._positive_control_well = positive_control_well
        self._sample_blow_height = sample_blow_height
        self._sample_bottom_height = sample_bottom_height
        self._sample_mix_vol = sample_mix_vol
        self._sample_mix_reps = sample_mix_reps
        self._sample_vol = sample_vol
        self._samples_per_cycle = int(math.ceil(samples_per_cycle / 8) * 8)
        self._source_plate_name = source_plate_name
        self._suck_height = suck_height
        self._suck_vol = suck_vol
        self._tempdeck_bool = tempdeck_bool
        self._tipracks_slots = tipracks_slots
        self._transfer_samples = transfer_samples
        self._tube_block_model = tube_block_model
        
        self._remaining_samples = self._num_samples
        self._samples_this_cycle = min(self._remaining_samples, self._samples_per_cycle)
    
    @property
    def num_cycles(self) -> int:
        return int(math.ceil(self._num_samples / self._samples_per_cycle))
    
    @labware_loader(0, "_source_plate")
    def load_source_plate(self):
        if self._transfer_samples:
            self._source_plate = self._ctx.load_labware('opentrons_96_aluminumblock_nest_wellplate_100ul', '5', self._source_plate_name)
        else:
            pass
    
    @labware_loader(1, "_tips20")
    def load_tips20(self):
        self._tips20 = [
            self._ctx.load_labware('opentrons_96_filtertiprack_20ul', slot)
            for slot in self._tipracks_slots
        ]
    
    @labware_loader(2, "_tips20_no_a")
    def load_tips20_no_a(self):
        self._tips20_no_a = [self._ctx.load_labware('opentrons_96_filtertiprack_20ul', '11', '20ul tiprack - no tips in row A')]
    
    @labware_loader(3, "_tips300")
    def loadtips300(self):
        self._tips300 = [self._ctx.load_labware('opentrons_96_filtertiprack_200ul', '10')]
    
    @labware_loader(4, "_tempdeck")
    def load_tempdeck(self):
        if self._tempdeck_bool:
            self._tempdeck = self._ctx.load_module('Temperature Module Gen2', '1')
        else:
            pass

    @labware_loader(5, "_pcr_plate")
    def load_pcr_plate(self):
        if self._tempdeck_bool:
            self._pcr_plate = self._tempdeck.load_labware('opentrons_96_aluminumblock_biorad_wellplate_200ul', 'PCR plate')
        else:
            self._pcr_plate = self._ctx.load_labware('opentrons_96_aluminumblock_biorad_wellplate_200ul', '1', 'PCR plate')
        
    @labware_loader(6, "_mm_strips")
    def load_mm_strips(self):
        self._mm_strips = self._ctx.load_labware('opentrons_96_aluminumblock_generic_pcr_strip_200ul', '4', 'mastermix strips')
    
    @labware_loader(7, "_tube_block")
    def load_tube_block(self):
        self._tube_block = self._ctx.load_labware(self._tube_block_model, '7', 'screw tube aluminum block for mastermix + controls')
    
    @instrument_loader(0, "_m20")
    def load_m20(self):
        self._m20 = self._ctx.load_instrument('p20_multi_gen2', 'right', tip_racks=self._tips20)
        self._m20.flow_rate.aspirate = self._aspirate_rate
        self._m20.flow_rate.dispense = self._dispense_rate
    
    @instrument_loader(0, "_p300")
    def load_p300(self):
        self._p300 = self._ctx.load_instrument('p300_single_gen2', 'left', tip_racks=self._tips300)
        self._p300.flow_rate.aspirate = self._aspirate_rate
        self._p300.flow_rate.dispense = self._dispense_rate
    
    @property
    def sources(self):
        if self._transfer_samples:
            return self._source_plate.rows()[0][:self.num_cols]
        else:
            return None
    
    @property
    def sample_dests(self):
        return self._pcr_plate.rows()[0][:self.num_cols]

    @property
    def control_dests(self):
        return self._pcr_plate.wells()[95]
    
    def _tipracks(self) -> dict:
        return {
            "_tips300": "_p300",
            "_tips20": "_m20",
            "_tips20_no_a": "_m20",
        }
    
    def pick_up_no_a(self):
        self.pick_up(self._m20, tiprack="_tips20_no_a")
        
    @property
    def mm_tubes(self):
        return self._tube_block.wells()[:1]

    @property
    def mm_tube_con(self):
        return self._tube_block.wells()[0]
    
    @property
    def mm_strips(self):
        return self._mm_strips.columns()[:1]
    
    @property
    def remaining_cols(self) -> int: 
        return int(math.ceil(min(self._remaining_samples, self._samples_per_cycle) / self._m20.channels))
    
    def fill_mm_strips(self):
        vol_per_strip_well = self.remaining_cols * self._mastermix_vol / len(self.mm_strips)
        
        has_tip = False        
        for j, (strip, tube) in enumerate(zip(self.mm_strips, self.mm_tubes)):
            for i, well in enumerate(strip):
                if self.run_stage("transfer mastermix {}/{} to strip {}/{}{}{}".format(i + 1, len(strip), j + 1, len(self.mm_strips), " " if self.num_cycles > 1 else "", self._cycle)):
                    if not has_tip:
                        self.pick_up(self._p300)
                        has_tip = True
                    self.logger.debug("filling mastermix at {}".format(well))
                    self._p300.transfer(vol_per_strip_well, tube.bottom(0.2), well, new_tip='never')

    def fill_control(self):
        if self.run_stage("Transfer mastermix to positive control hole in H12"):
            has_tip = True
            self._p300.transfer(self._mastermix_vol / self._mastermix_vol_headroom_aspirate, self.mm_tube_con.bottom(0.2),
                                self.control_dests, new_tip='never')
            if has_tip:
                self._p300.drop_tip()

    @property
    def mm_indices(self):
        return list(repeat(0, self._samples_per_cycle))
    
    def transfer_mm(self, stage="transfer mastermix {}/{}"):
        n = len(list(zip(self.mm_indices[::self._m20.channels], self.sample_dests[:self.remaining_cols])))
        for i, (m_idx, s) in enumerate(zip(self.mm_indices[::self._m20.channels], self.sample_dests[:self.remaining_cols])):
            if self.run_stage(stage.format(i + 1, n)):
                self.pick_up(self._m20)
                self._m20.transfer(self._mastermix_vol / self._mastermix_vol_headroom_aspirate, self.mm_strips[m_idx][0].bottom(0.8), s, new_tip='never')
                self._m20.drop_tip()
    
    def transfer_sample(self, vol: float, source, dest):
        self.logger.debug("transferring {:.0f} uL from {} to {}".format(vol, source, dest))
        self.pick_up(self._m20, tiprack="_tips20_no_a" if source.display_name.split(" ")[0] == self._positive_control_well else None)
        self._m20.transfer(vol, source.bottom(self._sample_bottom_height), dest.bottom(self._sample_bottom_height), new_tip='never')
        self._m20.mix(self._sample_mix_reps, self._sample_mix_vol, dest.bottom(self._sample_bottom_height))
        self._m20.blow_out(dest.top(self._sample_blow_height))
        self._m20.aspirate(self._suck_vol, dest.top(self._suck_height))  # suck in any remaining droplets on way to trash
        self._m20.drop_tip()
        
    def cycle_begin(self):
        self.logger.info(self.get_msg_format("current cycle", self._cycle.split(" ")[-1]))
        self._cycle = self._cycle if self.num_cycles > 1 else ""
    
    def run_cycle(self):
        self._cycle = self.stage
        n = self._remaining_samples
        self._samples_this_cycle = min(n, self._samples_per_cycle)
        
        self.cycle_begin()
        
        self.fill_mm_strips()
        self.fill_control()
        self.transfer_mm(stage="transfer mastermix to plate {}{}{}".format("{}/{}", " " if self.num_cycles > 1 else "", self._cycle))
        if self.sources is not None:
            for i, (s, d) in enumerate(zip(self.sources, self.sample_dests)):
                if self._transfer_samples and self.run_stage("transfer samples {}/{}".format(self._num_samples + min(self._m20.channels - self._remaining_samples, 0), self._num_samples)):
                    if self.num_cycles > 1:
                        self.msg_format("sample per cycle", (i + 1) * self._m20.channels, self._samples_this_cycle, self._cycle.split(" ")[-1])
                        self.logger.info(self.msg)
                    self.transfer_sample(self._sample_vol, s, d)
                    self.msg = ""
                self._remaining_samples -= self._m20.channels
                if self._remaining_samples <= 0 or n - self._remaining_samples >= self._samples_this_cycle:
                    break
        else:
            pass
    
    def body(self):
        self.logger.info(self.get_msg_format("number of cycles", self._num_samples, self.num_cycles))
        
        for i in range(self.num_cycles):
            self.run_stage("cycle {}/{}".format(i + 1, self.num_cycles))
            self.run_cycle()
            self.pause(self.get_msg_format("end of cycle", i + 1, self.num_cycles), color="yellow")
            if i + 1 < self.num_cycles and self.run_stage("pausing for next cycle {}/{}".format(i + 2, self.num_cycles)):
                self.pause("new cycle", color="green", blink=False, home=False)


if __name__ == "__main__":
    StationC(num_samples=480, metadata={'apiLevel': '2.3'}).simulate()


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
